#
# Licensed Materials - Property of IBM and/or HCL
# Rational ClearQuest
# (C) Copyright IBM Corp. 2017.  All Rights Reserved.
# (C) Copyright HCL Technologies Ltd. 2017  All Rights Reserved
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#
##############################################################################

# AS 6/16/17 - Adding support for Okta
# AS 7/11/17 - Fix issues with setting up for OIDC that were introduced for Okta work

use strict;
use warnings;
use CQPerlExt;
use Cwd;
use File::Copy;
use Switch;
use File::Spec;
use POSIX qw(strftime);
use Data::Dumper qw(Dumper);

select(STDERR); $| = 1;     #    temporal ordering.
select(STDOUT); $| = 1;     # (STDOUT must be last in this series.)

my $cqssosetup_version = "1.0.2";

#
# deal with Unix / Windows
#
my $Q = "\'";	# assume Unix
my $gIsWindows = 0;
my $FS = "/";
my $WSADMIN = "wsadmin.sh";
my $CQRPCEXE = "cqrpc";
my $MANAGEPROFILES = "manageProfiles.sh";
my $STARTSERVER = "startServer.sh";
my $STOPSERVER = "stopServer.sh";
my $DEFAULT_WAS_INSTALL_32 = "/opt/IBM/WebSphere/AppServer/";
my $DEFAULT_WAS_INSTALL_64 = $DEFAULT_WAS_INSTALL_32;

# constant used for specifying Jython over JACL
my $USE_JYTHON = 1;

# unfortunately, only available on Windows.
my $INSTALLUTIL = "installutil.exe";

if ($^O =~ /MSWin/)
{
	$gIsWindows = 1;
	$Q = "\"";
	$FS = "\\";
	$WSADMIN = "wsadmin.bat";
	$CQRPCEXE = "cqrpc.exe";
    $MANAGEPROFILES = "manageProfiles.bat";
    $STARTSERVER = "startServer.bat";
    $STOPSERVER = "stopServer.bat";
    $DEFAULT_WAS_INSTALL_32 = "C:\\Program Files (x86)\\IBM\\WebSphere\\AppServer\\";
    $DEFAULT_WAS_INSTALL_64 = "C:\\Program Files\\IBM\\WebSphere\\AppServer\\";
}

my $LOGSDIR = 'logs';
if (!(-d $LOGSDIR)) {
    mkdir $LOGSDIR;
}

my $bkts = strftime '%H%M%S_%Y%m%d', localtime;
my $logFileName = $LOGSDIR.$FS."cqssosetuplog.$bkts.txt";
my $LOGFILE;
open($LOGFILE, ">", $logFileName) or die "Could not open $logFileName for logging.\n";
print "Logging in $logFileName (some command output may be missing in log, use tee)\n";

sub tprint
{
    my $msg = shift;
    print $LOGFILE $msg;
    print STDOUT $msg;
}


my $bkpbase = "bkp";
my $bkpdir = $bkpbase.$FS.$bkts;
if (!(-d $bkpbase)) {
    tprint "Backup directory $bkpbase does not exist yet, creating it.\n";
    mkdir ($bkpbase);
} 
if (!(-d $bkpdir)) {
    tprint "Storing file backups in $bkpdir.\n";
    mkdir ($bkpdir);
} 

#
# Where is ClearQuest?
#
my $g_ClearQuestHomeDir = "";


# Some WAS stuff
#
# For wsadmin, JACL, the objects are "$AdminConfig" and "$AdminTask",
# When creating the command strings its convenient to just put $xxxxx but this
# would resolve to an undefined Perl variable. Define them here to 
# the right thing in the command string:
my $AdminConfig = "\$AdminConfig";
my $AdminTask = "\$AdminTask";
my $AdminApp = "\$AdminApp";

# TODO, have this initialized by config file?

my $g_configFileName = "ssoconfig.txt";
my $g_sso_config = ();
my $g_CQProfileName;
my $g_user;
my $g_password;

# What is the command we are running?
my $g_command;

sub logMsg
{
    my $msg = shift;
    if ($msg) {
        tprint "$msg\n";
    }    
}

sub runCmd
{
	my ($cmd) = @_;
	
	my $result;
	runningMsg("Running command: $cmd");
    eval {
        $result = system($cmd);
    };
    
    if ($@) {
        warnMsg("Error running command:\n$@\n");
        $result = -1;
    }
    if ($result) {
        my $error = $!;
        warnMsg("Warning, result code: $result\nError: $error\n");
    } else {
        runningMsg("Finished running command.");
    }
    return $result;
}

sub statusBanner
{
    my $status = shift;
    if ($status) {
        tprint "*************************************************************\n";
        tprint $status."\n";
        tprint "*************************************************************\n";
    }    
}

sub runningMsg
{
    my $msg = shift;
    if ($msg) {
        tprint "\n--- $msg\n\n";
    }    
}

sub progressMsg
{
    my $msg = shift;
    if ($msg) {
        tprint "\n+++ $msg\n\n";
    }    
}

sub warnMsg
{
    my $msg = shift;
    if ($msg) {
        tprint "\n!!! $msg\n\n";
    }    
}

sub errorMsg
{
    my $msg = shift;
    if ($msg) {
        tprint "\n*** $msg\n\n";
    }    
}

sub errorMsgAndDie
{
    my $msg = shift;
    errorMsg($msg);
    logMsg("!&! Unable to continue after previous error. Exiting...");
    die "Exiting...";
}

# Add quotes if any whitespace is detected in path.
sub quotePath
{
    my ($path) = @_;
    if ($path =~ /\s/g) {
        $path = "\"$path\"";
    }
    return $path;
}

# Takes an array of files and copies it to the backup folder.
sub copyToBackups
{
    foreach (@_) {
        copy($_, $bkpdir.$FS.$_);
    }
}

sub getFullPathClearQuest
{
    # Have we already found it?
    return $g_ClearQuestHomeDir unless (!$g_ClearQuestHomeDir);
    
    if ($gIsWindows && $g_sso_config->{SSO_CQ_HOME_WIN}) {
        $g_ClearQuestHomeDir = $g_sso_config->{SSO_CQ_HOME_WIN};
    } elsif (!$gIsWindows && $g_sso_config->{SSO_CQ_HOME_UNIX}) {
        $g_ClearQuestHomeDir = $g_sso_config->{SSO_CQ_HOME_UNIX};
    } else {
        $g_ClearQuestHomeDir = $ENV{'CLEARQUEST_HOME'};	# on Windows
        if (!defined($g_ClearQuestHomeDir) or $g_ClearQuestHomeDir eq "")
        {
        	# don't give up just yet
        	$g_ClearQuestHomeDir = $ENV{'CQ_HOME'};	# on Unix
        	if (!defined($g_ClearQuestHomeDir) or $g_ClearQuestHomeDir eq "")
        	{
        		# keep trying, one last time ...
        		$g_ClearQuestHomeDir = $ENV{'RATIONAL_HOME'};
        		if (defined($g_ClearQuestHomeDir) and $g_ClearQuestHomeDir ne "")
        		{
        			# 'RATIONAL_HOME' doesn't have 'ClearQuest"
        			$g_ClearQuestHomeDir .= $FS."ClearQuest";
        		}
        	}
        }
    }
    if (substr($g_ClearQuestHomeDir, -1) ne $FS)
    {
    	$g_ClearQuestHomeDir .= $FS;
    }
    if (-d $g_ClearQuestHomeDir) {
        return $g_ClearQuestHomeDir;
    } else {
        errorMsgAndDie("Could not find ClearQuest home directory: $g_ClearQuestHomeDir");
    }
}

# Path to the cqrpc, but not including the cqrpc itself. Has a FS at the end.
my $g_fullPathToCqRpc;
sub getFullPathToCqRpc
{
    return $g_fullPathToCqRpc unless (!$g_fullPathToCqRpc);
    
    # TODO - make it tolerant to a missing FS at the end.
    if ($gIsWindows && $g_sso_config->{SSO_CQ_CQRPC_WIN}) {
        $g_fullPathToCqRpc = $g_sso_config->{SSO_CQ_CQRPC_WIN};
    } elsif (!$gIsWindows && $g_sso_config->{SSO_CQ_CQRPC_UNIX}) {
        $g_fullPathToCqRpc = $g_sso_config->{SSO_CQ_CQRPC_UNIX};
    } 
    if (!$g_fullPathToCqRpc) {
        $g_fullPathToCqRpc = getFullPathClearQuest();
        $g_fullPathToCqRpc .= "cqweb".$FS."bin".$FS;
    }
    if (substr($g_fullPathToCqRpc, -1) ne $FS)
    {
    	$g_fullPathToCqRpc .= $FS;
    }
    if (-d $g_fullPathToCqRpc) {
        return $g_fullPathToCqRpc;
    } else {
        errorMsgAndDie("Error, could not find CQRPC directory: $g_fullPathToCqRpc");
    }
}

sub getFullPathToInstallUtilCmd
{
    return getFullPathClearQuest().$INSTALLUTIL;
}

# For finding wsadmin, maybe its okay if its the profile itself.
my $g_WasPath;
sub findWASAppServerDir
{
    return $g_WasPath unless (!$g_WasPath);

    $g_WasPath = $g_sso_config->{SSO_WAS_PATH} || "";
    if (!$g_WasPath) {
        # Try default install path?
        if (-d $DEFAULT_WAS_INSTALL_32) {
            $g_WasPath = $DEFAULT_WAS_INSTALL_32;
        } elsif (-d $DEFAULT_WAS_INSTALL_64) {
            $g_WasPath = $DEFAULT_WAS_INSTALL_64;
        }
    }
    
    # TODO, make this tolerant to a missing FS in the path 
    # when it is not there we need to add it.

    if ($g_WasPath && -d $g_WasPath) {
        return $g_WasPath;
    } else {
        errorMsgAndDie("Could not find WebSphere path, tried: \"$g_WasPath\"");
    }
}

my $g_WasBinPath;
sub findWASAppServerDirBin
{
    return $g_WasBinPath unless (!$g_WasBinPath);

    $g_WasBinPath = findWASAppServerDir()."bin".$FS;
    if (-d $g_WasBinPath) {
        return $g_WasBinPath;
    } else {
        errorMsgAndDie("Could not find WAS bin path, tried: \"$g_WasBinPath\"");
    }
}

my $g_profilePath;
# Returns the path to profile, no FS at the end.
sub getProfilePath
{
    if (!$g_profilePath) {
        my $maybeProfilePath = $g_sso_config->{SSO_WAS_PROFILE_PATH};
        if (!$maybeProfilePath) {
            if ($g_CQProfileName eq "cqwebprofile") {
                # this might be the default install where IM creates the cqwebprofile.
                # Try to find the profile path assuming that its the default CQ install profile.
                $maybeProfilePath = getFullPathClearQuest();
                $maybeProfilePath .= "cqweb".$FS."cqwebprofile";
            } else {
                my $manageprofCmd = getMangeProfilesCmd();
                $manageprofCmd .= " -getPath -profileName $g_CQProfileName";
                progressMsg("Trying to get profile with this: $manageprofCmd");
                $maybeProfilePath = `$manageprofCmd`;
                chomp($maybeProfilePath);
            }
        }
        if (-d $maybeProfilePath) {
            $g_profilePath = $maybeProfilePath;
            # remove any trailing file separator
            $g_profilePath =~ s/$FS$//;
            progressMsg("Profile path set to $g_profilePath");
        } else {
            errorMsgAndDie("Could not find CQ Web profile path, \"$maybeProfilePath\". Please set a valid profiile path in configuration file.");
        }
    }
    return $g_profilePath;
}

my $g_ProfileBin;
sub getProfileBin
{
    return $g_ProfileBin unless (!$g_ProfileBin);

    $g_ProfileBin = getProfilePath().$FS."bin".$FS;
    if (-d $g_ProfileBin) {
        return $g_ProfileBin;
    } else {
        errorMsgAndDie("Could not find Profile bin path, tried: $g_ProfileBin");
    }
}

sub getNodeFromProfile
{
    my $node;
    my $wsadmin = getWsAdminCmd();
    my $cmd = $wsadmin." -profileName $g_CQProfileName -user $g_user -password $g_password -lang jython ".
            "-c \"print 'Node='+AdminControl.getNode()\"";
    runningMsg("Running command to get profile node: $cmd");
    open (GETNODE, "-|", $cmd);
    while (<GETNODE>) {
        logMsg($_);
        if (/Node=(.+)/) {
            $node = $1;
            last;
        }
    }
    close GETNODE;
    progressMsg("Got profile node: $node");
    return $node;
}

sub getWASNode
{
    my ($configName, $wasEnvVar) = @_;
    if (!$g_sso_config->{SSO_WAS_NODE}) {
        $g_sso_config->{SSO_WAS_NODE} = getNodeFromProfile();
        if (!$g_sso_config->{SSO_WAS_NODE})
        {
            if ($ENV{$wasEnvVar}) {
                $g_sso_config->{SSO_WAS_NODE} = $ENV{WAS_NODE};
            } else {
                errorMsgAndDie("Could not determine WAS_NODE. Try running WAS_PROFILE/bin/setupCmdLine first.");
            }
        }
    }
    return $g_sso_config->{SSO_WAS_NODE};
}

sub getWASServerName
{
    my ($configName, $wasEnvVar) = @_;
    if (!$g_sso_config->{SSO_WAS_SERVER_NAME}) {
        if ($g_CQProfileName eq "cqwebprofile") {
            $g_sso_config->{SSO_WAS_SERVER_NAME} = "server1";
        } else {
            errorMsgAndDie("Could not determine server name. Specify this as SSO_WAS_SERVER_NAME in the configuration file.");
        }
    }
    return $g_sso_config->{SSO_WAS_SERVER_NAME};
}

sub findOIDCInstallScript
{
    return findWASAppServerDirBin()."installOIDCRP.py";
}

sub findOIDCSetupScript
{
    return "setupoidc.py";
}

sub findSAMLACSInstallScript
{
    return findWASAppServerDirBin()."installSamlACS.py";
}

sub getWsAdminCmd
{
    return quotePath(findWASAppServerDir()."bin".$FS.$WSADMIN);
}

sub getMangeProfilesCmd
{
    return quotePath(findWASAppServerDir()."bin".$FS.$MANAGEPROFILES);
}

sub getStartServerCmd
{
    return quotePath(getProfileBin().$STARTSERVER);
}

sub getStopServerCmd
{
    return quotePath(getProfileBin().$STOPSERVER);
}

sub stringFromTemplateAndHash
{
    my ($propText, $propsHash) = @_;
    for (keys %$propsHash) {
        my $propKey = $_;
        my $subValue = $propsHash->{$propKey} || "";
        $propText =~ s/\%$propKey\%/$subValue/ig;
    }
    return $propText;
}

sub writeFileFromTemplateAndHash
{
    my ($filename, $propText, $propsHash) = @_;
    
    $propText = stringFromTemplateAndHash($propText,$propsHash);

    my $PROPFILE;
    open(PROPFILE, ">", $filename) or die "Could not open $filename for writing.\n";
    print PROPFILE $propText;
    close(PROPFILE);
}

sub restart_server
{
    #stopSever:
    #ADMU4000I: Server server1 stop completed.
    #if already stopped then:
    #ADMU0509I: The server "server1" cannot be reached. It appears to be stopped.
    #startServer:
    #ADMU3000I: Server server1 open for e-business; process id is 5304
    my $startServerCmd = getStartServerCmd();
    my $stopServerCmd = getStopServerCmd();
    my $serverName = getWASServerName();
    
    progressMsg("Stopping $serverName...");    
    my $stopServerResp = `$stopServerCmd $serverName -user $g_user -password $g_password `;
    print "$stopServerResp\n";
    if ($stopServerResp =~ /ADMU4000I/i || $stopServerResp =~ /ADMU0509I/i) {
        # stop was successful, now start:
        progressMsg("Stop was successful, now restarting $serverName...");    
        my $startServerResp = `$startServerCmd $serverName `;
        tprint "$startServerResp\n";
        if ($startServerResp =~ /ADMU3000I/i) {
            progressMsg("Restart was successful.");
        } else {
            warnMsg("Restart was not successful.");
        }
    } else {
        warnMsg("Could not restart $serverName. Try restarting $serverName manually.");
    }
}


my $OIDC_RP_JSON_TEMPLATE = <<EOF;
{
   "client_secret":"%SSO_OIDC_RP_SECRET%",
   "client_id":"%SSO_OIDC_RP_CLIENTNAME%",
   "client_name":"%SSO_OIDC_RP_CLIENTNAME%",
   "token_endpoint_auth_method":"client_secret_basic",
   "preauthorized_scope" : "openid profile email general phone address",
   "scope" : "openid profile scope1 email general phone address",
   "grant_types":[
      "authorization_code",
      "client_credentials",
      "implicit",
      "refresh_token",
      "urn:ietf:params:oauth:grant-type:jwt-bearer"
   ],
   "response_types":[
      "code",
      "token",
      "id_token token"
   ],
   "application_type":"web",
   "subject_type":"public",
   "introspect_tokens":true,
   "redirect_uris":[
      "https://%SSO_CQWEB_HOST%:%SSO_CQWEB_PORT%/oidcclient/%SSO_OIDC_RP_ID%"
   ],
   "trusted_uri_prefixes":[
      "https://%SSO_CQWEB_HOST%:%SSO_CQWEB_PORT%/cqweb"
   ]
}
EOF

my $OIDC_RP_SERVERXML_TEMPLATE = <<EOF;
            <client name="%SSO_OIDC_RP_CLIENTNAME%" secret="%SSO_OIDC_RP_SECRET%"
                displayname="%SSO_OIDC_RP_CLIENTNAME%"
                redirect="https://%SSO_CQWEB_HOST%:%SSO_CQWEB_PORT%/oidcclient/%SSO_OIDC_RP_ID%"
                scope="openid profile scope1 email phone address"
                preAuthorizedScope="openid"
                enabled="true"/>
EOF

my $OIDC_REGISTER_INSTRUCTIONS = <<EOF;    
    You must register your CQ Web Relying Party on your OIDC IDP Server.
    Consult documentation of your OIDC IDP Server for how to register
    a relying party.
    Two files have been generated to assist with this. The JSON file can
    be used for REST based registration, or for Jazz Authorization 
    Server (JAS). The XML file has snippet you can use for a WebSphere
    Liberty server configured as an OIDC IDP server that use a "LocalStore"
    for registering relying parties. 
    
    If you are using a Jazz Authorization Server (JAS) for your OIDC IDP, 
    then copy the file "cqwebreg.json" to your the machine hosting
    your JAS and run the following command
    <JAS_PATH>\\cli\\ldclient.bat -c -u <JAS_ADMIN_USER>:<JAS_ADMIN_PASSWORD> cqwebreg.json
    
    If the OIDC IDP supports dynamic client registration, you can also 
    a REST API to register the CQ Web relying party.
    Detailed instructions for this may be found here:
    https://openid.net/specs/openid-connect-registration-1_0.html
    You can use the "cqwebreg.json" as the content for your client
    registration.
    
EOF

sub idp_post_setup_oidc
{
    # Can't really setup the OIDC IDP here. We can export a JSON file for use in JAS ldclient 
    # and an XML fragment for the Liberty server.xml file.
    tprint $OIDC_REGISTER_INSTRUCTIONS;
    writeFileFromTemplateAndHash("cqwebreg.json", $OIDC_RP_JSON_TEMPLATE, $g_sso_config);
    writeFileFromTemplateAndHash("cqwebreg.xml", $OIDC_RP_SERVERXML_TEMPLATE, $g_sso_config);
}

my $SAML2_REGISTER_INSTRUCTIONS = <<EOF;    

    The metadata for the CQ Web service provider (SP) has been exported to
    the current directory as "cqwebsamlspdata.xml" 
    You must import this at the SAML2 identity provider so that it trusts 
    this SP. Follow the instructions of the identity provider for importing
    service provider metadata.

EOF

sub idp_post_setup_saml2
{
    my $cqwebsamspdataFileName = File::Spec->rel2abs("cqwebsamlspdata.xml");
    progressMsg("Exporting SP metadata to $cqwebsamspdataFileName");
    # must remove existing first.
    unlink $cqwebsamspdataFileName;
    my @commands = (
        "AdminTask.exportSAMLSpMetadata('-spMetadataFileName $cqwebsamspdataFileName -ssoId 1')",
        "AdminConfig.save()"
    );
    runWsAdminCommands(\@commands, $USE_JYTHON );
    
    # Print post-setup instructions.
    tprint $SAML2_REGISTER_INSTRUCTIONS;
}

# This doesn't do any setup, it just creates support files you may need to 
# do any manual post setup steps.
sub post_setup_steps
{
    statusBanner("Post setup steps.");
    my $ssoType = $g_sso_config->{SSO_TYPE};

    switch ($ssoType) {
        case "OIDC" {
            idp_post_setup_oidc();
        }
        case "SAML2" {
            idp_post_setup_saml2();
        }
        else {
            errorMsgAndDie("Unknown SSO Type specified: \"$ssoType\"");
        }
    }
}

sub sso_enable
{
    my $doCqrpc = 1;
    my $doMasterDbs = 1;
    my $doCqWeb = 1;
    my $doWebServer = 1;
    my $doRestart = 1;
    my $doPostSetupSteps = 1;
    my $doIDPSetup = 1;
    
    my @unusedArgs = ();
        
    while (my $extraOp = shift @ARGV) {
        $extraOp = lc($extraOp);
        if ($extraOp eq "-config") {
            $g_configFileName = shift(@ARGV);
        } elsif ($extraOp eq "-nocqrpc") {
            $doCqrpc = 0;
        } elsif ($extraOp eq "-nomasterdbs") {
            $doMasterDbs = 0;
        } elsif ($extraOp eq "-nocqweb") {
            $doCqWeb = 0;
        } elsif ($extraOp eq "-nowebserver") {
            $doWebServer = 0;
        } elsif ($extraOp eq "-norestart") {
            $doRestart = 0;
        } elsif ($extraOp eq "-nopost") {
            $doPostSetupSteps = 0;
        } else {
            push @unusedArgs, $extraOp;
        }
    }
    if (($doCqWeb || $doMasterDbs) && !$doCqrpc) {
        warnMsg("cqweb and masterdbs needs cqrpc setup. Doing cqrpc setup.");
        $doCqrpc = 1;
    }
    
    # only do restart if we didn't have the -norestart and we did at least some of the
    # SSO enablement.
    $doRestart = $doRestart && ($doCqrpc || $doMasterDbs || $doCqWeb || $doWebServer);
    
    progressMsg("Enabling CQ Web server for SSO.");
    
    my $enable = 1;
    
    read_configuration();
    if ($doCqrpc) {
        setup_cqrpc($enable);
    }
    if ($doMasterDbs) {
        setup_masterdbs($enable);
    }
    if ($doCqWeb) {
        setup_cqweb();
    }
    if ($doWebServer) {
        setup_webserver($enable);
    }
    if ($doRestart) {
        restart_server($enable);
    }
    if ($doPostSetupSteps) {
        post_setup_steps($enable);
    }

}

sub sso_disable
{
    my $doCqrpc = 1;
    my $doMasterDbs = 1;
    my $doCqWeb = 1;
    my $doWebServer = 1;
    my $doRestart = 1;
    my $doPostSetupSteps = 1;
    my $doIDPSetup = 1;
    
    my @unusedArgs = ();
        
    while (my $extraOp = shift @ARGV) {
        $extraOp = lc($extraOp);
        if ($extraOp eq "-config") {
            $g_configFileName = shift(@ARGV);
        } elsif ($extraOp eq "-nomasterdbs") {
            $doMasterDbs = 0;
        } elsif ($extraOp eq "-nocqweb") {
            $doCqWeb = 0;
        } elsif ($extraOp eq "-nowebserver") {
            $doWebServer = 0;
        } elsif ($extraOp eq "-norestart") {
            $doRestart = 0;
        } else {
            push @unusedArgs, $extraOp;
        }
    }
    
    # only do restart if we didn't have the -norestart and we did at least some of the
    # SSO enablement.
    $doRestart = $doRestart && ($doCqrpc || $doMasterDbs || $doCqWeb || $doWebServer);
    
    progressMsg("Disabling CQ Web server for SSO.");
    
    read_configuration();

    my $disable = 0;
    
    if ($doMasterDbs) {
        setup_masterdbs($disable);
    }
    if ($doCqWeb) {
        disable_cqweb();
    }
    if ($doWebServer) {
        setup_webserver($disable);
    }
    if ($doRestart) {
        restart_server($disable);
    }
}

sub sso_restart
{
    read_configuration();
    restart_server();
}


#Example config:
# See ssoconfig.txt
#
sub read_configuration
{
    statusBanner("Reading SSO configuration from $g_configFileName");
   	if (open(FH, "<", $g_configFileName)) {
   	    while (<FH>) {
    		s/\r\n//;
    		chomp;
    		if (!$_ or $_ =~/^\s*#/ or $_ =~/^\s*$/) {
   	            # blank lines, comments, ignore.
   	            next;
   	        }
   	        if (/(.+)=(.*)/) {
   	            print "value '$1' = '$2'\n";
   	            $g_sso_config->{$1} = $2;
   	        } else {
   	            warnMsg("Warning, could not parse line: $_\n");
   	        }
   	    }
   	    close(FH);
   	} else {
   	    errorMsgAndDie("Can't open file $g_configFileName");
   	}
   	
   	$g_CQProfileName = $g_sso_config->{SSO_WAS_PROFILE} || "cqwebprofile";
    $g_user = $g_sso_config->{SSO_WAS_ADMIN_USER};
    $g_password = $g_sso_config->{SSO_WAS_ADMIN_PASSWORD};
    
   	
   	statusBanner("Done reading SSO configuration");
}

# Run the cqrpc 
# see http://www.ibm.com/support/knowledgecenter/SSSH5A_8.0.1/com.ibm.rational.clearquest.webadmin.doc/topics/t_smartc_sso_conf_cmsvr.htm
#
sub setup_cqrpc
{
    statusBanner("Generating SSO password for cqrpc");

    my $toCqrpcPath = getFullPathToCqRpc();    
    # on Linux, need to add cqrpc's directory to the path so it can pick up
    # its libraries.
    if (!$gIsWindows) {
        $ENV{LD_LIBRARY_PATH} .= ":$toCqrpcPath";
    } else {
        # need to put ClearQuest on the path
        my $cqPath = getFullPathClearQuest();
        $ENV{PATH} = "$cqPath;".$ENV{PATH};
        print "Path = $ENV{PATH}\n";
    }
    my $cmd = quotePath($toCqrpcPath.$CQRPCEXE)." -sso_password ".$g_sso_config->{SSO_PASSWORD};
    if (runCmd($cmd)) {
        die "Could not generate encrypted password.\n";
    }
    my $FH;
    open (FH, "<", "sso.properties.temp");
    $g_sso_config->{SSO_PASSWORD_ENCR} = <FH>;
    close(FH);
    
    # Copy sso.properties (with just FIPS pw in the same directory as cqrpc. )
    # Note, Perl File::Copy doesn't like quotes in the destination so we don't enclose the
    # destination in quotes.
    my $dest = $toCqrpcPath."sso.properties";
    progressMsg("Copying sso.properties.temp to $dest\n");
    if (!copy("sso.properties.temp", $toCqrpcPath."sso.properties")) {
        die "Copy failed $!\n";
    }
    #if (!copy("sso.properties", quotePath($toCqrpcPath)) {
    #    die "Copy failed $!\n";
    #}
    statusBanner("Done generating SSO password for cqrpc");
}

sub use_sso_apis
{
    return (has_sso_apis(@_) && !$g_sso_config->{SSO_CQ_USE_INSTALLUTIL})
}

sub has_sso_apis
{
    my $user = shift;
    my $pw = shift;
    my $dbset = shift;
    my $hasapis = 0;
    if ($g_sso_config->{SSO_CQ_HAS_SSO_APIS}) {
        $hasapis = 1;
    } else {
        progressMsg("Checking for existence of SSO APIs");
    
        
    	my $session = CQAdminSession::Build();
    	eval {
    	    $session->Logon($user, $pw, $dbset);
    	}; 
    	die "Cannot logon to dbset $dbset: $@\n" if ($@);
    	
    	my $ssoSetting;
        eval {
            $ssoSetting = $session->IsSSOEnabled();
        };
        if ($@ =~ /Undefined subroutine/ig) {
            $hasapis = 0;
        } else {
            # If there was an unexpected error checking API then die:
            die "Could not check SSO API existence. Encountered unexpected error: $@" if ($@);
            $hasapis = 1;        
        }
        $g_sso_config->{SSO_CQ_HAS_SSO_APIS} = $hasapis;
        progressMsg($hasapis ? "SSO APIs exist." : "Could not find SSO APIs");
    }
    return $hasapis;
}

sub setup_masterdbs_using_apis
{
    my $user = shift;
    my $pw = shift;
    my $dbset = shift;
    my $ssopw = shift;
    my $enable = shift;
    
	my $session = CQAdminSession::Build();
	eval {
	    $session->Logon($user, $pw, $dbset);
	}; 
	if ($@) {
	    errorMsgAndDie("Cannot logon to dbset $dbset: $@");
	}
	
	eval {
	    if ($enable) {
    	    $session->ConfigureSSO(1, $ssopw);    	
    	} else {
    	    $session->ConfigureSSO(0, "");
    	}
	};
	if ($@) {
	    errorMsgAndDie("Cannot enable SSO on $dbset: $@");
	}
	my $ssoEnabled;
    eval {
        $ssoEnabled = $session->IsSSOEnabled();
    };
    if ($ssoEnabled != $enable) {
	    errorMsgAndDie("Cannot ".($enable ? "enable" : "disable")." SSO on $dbset for unknown reasons");    
    }
    progressMsg("SSO configuration on $dbset was successful.");
}

sub setup_masterdbs_using_installutil
{
    my $user = shift;
    my $pw = shift;
    my $dbset = shift;
    my $ssopw = shift;
    my $enable = shift;
    
    # just for password, if its empty, use empty quotes.
    $pw = ($pw) ? $pw : "\"\"";

    # enbable sso
    my $installUtilCmd = getFullPathToInstallUtilCmd();
    my $iucmd = $enable ? "enablessologin" : "disablessologin";
    my $cmd = quotePath($installUtilCmd)." $iucmd $dbset $user $pw";
    runCmd($cmd);
    
    if ($enable) {
        $cmd = quotePath($installUtilCmd)." setssoclientpassword $dbset $user $pw";
        runningMsg("Running command: $cmd\n");
        if (!open(INSTALLUTILPIPE, "| $cmd")) {
            errorMsgAndDie("Could not start installutil command: $!.");
        }
        # It will ask for SSO password, nothing more.
        print INSTALLUTILPIPE "$ssopw\n";
        close(INSTALLUTILPIPE);
    }
}


sub verifyEnoughForDbSet
{
    if (scalar(@_) % 3) {
        warnMsg("Warning, incorrect number of strings in SSO_CQ_DATABASE_INFO. ".
            "A multiple of 3 is expected since it must include 'dbset,user,pw' ".
            "for each dbset. THe following incomplete database info was ignored: ".
            join(",",@_));
        return 0;
    }
    return 1;
}

# returns an array of values, dbset,user,pw,dbset2,user2,pw2, etc... where
# the array should be 0 or a multiple of 3, but we don't check here.
sub getDbInfo
{
    return split(",", $g_sso_config->{SSO_CQ_DATABASE_INFO} || "");
}

# Setup the master dbs with the SSO password, can only do this on Windows now.
#
sub setup_masterdbs
{
    my $enable = shift;
    if ($g_sso_config->{SSO_CQ_DATABASE_INFO}) {
        my @dbinfo = getDbInfo();
        
        if (scalar(@dbinfo) % 3) {
            warnMsg("Warning, incorrect number of strings in SSO_CQ_DATABASE_INFO. ".
                "A multiple of 3 is expected since it must include 'dbset,user,pw' ".
                "for each dbset.");
        }
        
        while (scalar(@dbinfo)) {
            if (verifyEnoughForDbSet(@dbinfo)) {
                my $dbset = shift @dbinfo;
                my $user = shift @dbinfo;
                my $pw = shift @dbinfo;
                my $ssopw = $g_sso_config->{SSO_PASSWORD};
                # remove any quotes if there.
                $dbset =~ s/\"(\.*)\"/$1/;
                $user =~ s/\"(\.*)\"/$1/;
                $pw =~ s/\"(\.*)\"/$1/;
                

                if (use_sso_apis($user, $pw, $dbset)) {
                    setup_masterdbs_using_apis($user, $pw, $dbset, $ssopw, $enable);
                } else {
                    setup_masterdbs_using_installutil($user, $pw, $dbset, $ssopw, $enable);
                }

            } else {
                last;
            }
        }
    } else {
        warnMsg("Warning, no databases will be configured for SSO. ".
            "Use the SSO_CQ_DATABASE_INFO configuration to specify databases.\n");
    }
}

my $WEB_XML_SSO = <<EOF;    
    <!-- Security-constraint for resources that should be protected by SSO. -->
    <security-constraint>
        <web-resource-collection>
            <web-resource-name>secure</web-resource-name>
            <url-pattern>/*</url-pattern>            
        </web-resource-collection>          
        <auth-constraint>
            <role-name>ClearQuestUsers</role-name>
        </auth-constraint>          
        <user-data-constraint>
            <transport-guarantee>CONFIDENTIAL</transport-guarantee>
        </user-data-constraint>
    </security-constraint>
        
    <!-- Security-constraint for resources that should not be protected by SSO. -->
    <!-- These are protected by CQ security. -->
    <security-constraint>
        <web-resource-collection>
            <web-resource-name>CQSafe</web-resource-name>
            %SSO_CQWEB_OSLC_CONSTRAINT%
            <url-pattern>/oauth-request-consumer/*</url-pattern>
            <url-pattern>/oauth-authorize-consumers/*</url-pattern>
            <url-pattern>/oauth-request-token/*</url-pattern>
            <url-pattern>/oauth-authorization/*</url-pattern>
            <url-pattern>/oauth-access-token/*</url-pattern>
            <url-pattern>/scripts/*</url-pattern>
            <url-pattern>/images/*</url-pattern>
            <url-pattern>/stylesheets/*</url-pattern>
            <url-pattern>/gadgets/*</url-pattern>
            <url-pattern>/_rcm-resources/*</url-pattern>
            <url-pattern>/cqquerywizard.cq</url-pattern>
            <url-pattern>/cqartifactdetails.cq</url-pattern>
            <url-pattern>/cqqueryresults.cq</url-pattern>
        </web-resource-collection>
        <user-data-constraint>
            <transport-guarantee>NONE</transport-guarantee>
        </user-data-constraint>
    </security-constraint>

    <!-- Security-role for certificate authentication or any SSO. -->
    <security-role>
        <role-name>ClearQuestUsers</role-name>
    </security-role> 
EOF

my $WEB_XML_CLIENT_CERT = <<EOF;
    <!-- Login config for client certificates -->
    <login-config>
        <auth-method>CLIENT-CERT</auth-method>
    </login-config>    
EOF

my $WEB_XML_SSO_BEGIN = "<!-- !!! CQSSOSETUP_BEGIN - DO NOT EDIT MANUALLY USE cqssosetup.pl tool to remove -->\n";
my $WEB_XML_SSO_END   = "<!-- !!! CQSSOSETUP_END - DO NOT EDIT MANUALLY USE cqssosetup.pl tool to remove -->\n";

my $SSO_PROP_TEMPLATE = <<EOF;
#
# Licensed Materials - Property of HCL
# Rational ClearQuest 
# (C) Copyright HCL Technologies Ltd. 2016  All Rights Reserved
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp or HCL.
#
#                            sso.properties
#

################################################################################
### Properties used by Single Sign-On feature of ClearQuest Web             
### 
### Parameter names and values are case-sensitive.
###
################################################################################


### These parameters are used by the CQ Web Server machine.

### Parameter: SSO_CLIENT_ENABLE
### Valid Values: 
###		true or false
###
### Enables the Single Sign-On feature for ClearQuest Web machine.
###
### Default Value: false
###
SSO_CLIENT_ENABLE=%SSO_CLIENT_ENABLE%


### Parameter: SSO_CLIENT_PASSWORD
### Valied Values:
###		The encrypted password produced from the SSO password utility.
###
### Default Value: N/A
###
SSO_CLIENT_PASSWORD=%SSO_CLIENT_PASSWORD%


### Parameter: SSO_LOGIN_MODE
### Specifies how SSO is accomplished.
### Valied Values:
###     COOKIE      Use a cookie value for login credentials
###     HEADER      Use an HTTP header value for login credentials
###     REQUEST     Use the getRemoteUser() method of the HttpServletRequest to retrieve login credentials
###     OIDC        Use OpenID Connect provider
###     SAML2       Use a SAML2 identity provider
### Default Value: HEADER
###
SSO_LOGIN_MODE=%SSO_LOGIN_MODE%


### Parameter: SSO_LOGIN_USERNAME_IDENTIFIER
### Valid Values:
###		A value specifying where the CQ username is stored.
###
###	This value is used to to retrieve the login name for sign-on.
###
### Default Value: N/A
###
SSO_LOGIN_USERNAME_IDENTIFIER=%SSO_LOGIN_USERNAME_IDENTIFIER%

### Parameter: SSO_ENABLE_USER_PROFILE_MODIFY
### Valid Values: 
###		true or false
###
### Specifies whether users are allowed to modify their user profiles
### in their web sessions.
###
### Default Value: false
###
SSO_ENABLE_USER_PROFILE_MODIFY=%SSO_ENABLE_USER_PROFILE_MODIFY%

### Parameter: SSO_USES_JAS_OIDC
### Valid Values: 
###     true or false
###
### Specifies whether the SSO login mode uses a Jazz Authorization Server (JAS)
### and OpenID Connect (OIDC). Set this to true if ClearQuest uses the JAS
### as an OIDC provider.
###
### Default Value: false
###
SSO_USES_JAS_OIDC=%SSO_USES_JAS_OIDC%
EOF

sub getSSOPropertiesHash
{
    my $enabled = ($g_command eq "enable");
    
    my $propHash = {
        SSO_CLIENT_ENABLE =>  $enabled ? "true" : "false",
        SSO_CLIENT_PASSWORD => $enabled ? $g_sso_config->{SSO_PASSWORD_ENCR} : "",
        SSO_LOGIN_MODE => $enabled ? getSSOLoginModeForSSOProps() : "HEADER",
        SSO_LOGIN_USERNAME_IDENTIFIER => "",
        SSO_ENABLE_USER_PROFILE_MODIFY => getSSOAllowProfileModify(),
    };
    
    # add the g_sso_config hash in too.
    @{$propHash}{keys %$g_sso_config} = values %$g_sso_config;
    
    return $propHash;
}

#TODO  - check for invalid sso type?
sub getSSOLoginMode
{
    my $ssoType = $g_sso_config->{SSO_TYPE};
    if ($ssoType eq "LTPA2" || $ssoType eq "CERT") {
        $ssoType = "REQUEST";
    }
    return $ssoType;
}

sub getSSOLoginModeForSSOProps
{
    return $g_sso_config->{SSO_LOGIN_MODE_OVERRIDE_FOR_PROPS} || getSSOLoginMode();
}

sub writeSSOProperties
{
    writeFileFromTemplateAndHash("sso.properties", $SSO_PROP_TEMPLATE, getSSOPropertiesHash());
}

# TODO make this a config setting.
sub getSSOAllowProfileModify
{
    return "false";
}

sub getCQWebFilesToEdit
{
    my $profilePath = getProfilePath();
    runWsAdminFile("getfiles.py", quotePath($profilePath));
}

sub getDbSets
{
    my @dbsets = ();
    if ($g_sso_config->{SSO_CQ_DATABASE_INFO}) {
        my @dbinfo = getDbInfo();
        
        while (scalar(@dbinfo)) {
            if (verifyEnoughForDbSet(@dbinfo)) {
                my $dbset = shift @dbinfo;
                my $user_dontcare = shift @dbinfo;
                my $pw_dontcare = shift @dbinfo;
                $dbset =~ s/\"(\.*)\"/$1/;
                push @dbsets, $dbset;
            } else {
                last;
            }
        }
    }
    return @dbsets;
}
sub getOslcConstraint
{
    my $configOslcConstraint = shift || "";
    if ($g_sso_config->{SSO_USES_JAS_OIDC} && $g_sso_config->{SSO_USES_JAS_OIDC} eq "true") {
        # for JAS/OIDC SSO, need to add the discovery URLs, e.g.
        # and for the gadgets too.
        # <url-pattern>/oslc/repo/<dbset>/discovery</url-pattern>
        foreach (getDbSets()) {
            $configOslcConstraint .= "<url-pattern>/oslc/repo/$_/discovery</url-pattern>";
            $configOslcConstraint .= "<url-pattern>/oslc/repo/$_/queryGadget</url-pattern>";
            $configOslcConstraint .= "<url-pattern>/oslc/repo/$_/chartGadget</url-pattern>";
            $configOslcConstraint .= "<url-pattern>/oslc/repo/$_/gadgets/*</url-pattern>";
        }
    }
    return $configOslcConstraint;
}


#http://www.ibm.com/support/knowledgecenter/SSSH5A_8.0.1/com.ibm.rational.clearquest.webadmin.doc/topics/t_smartc_sso_conf_cqw_client.htm
sub setup_cqweb
{
    statusBanner("Setting up CQ WAS Server for SSO");
    
    #remove these before getting copies from the profile.
    unlink("sso.properties");
    unlink("web.xml");
    
    progressMsg("Getting current web.xml and sso.properties and backing them up");
    getCQWebFilesToEdit();
    
    # make backups
    copyToBackups("sso.properties", "web.xml");
    
    # For web.xml add the $WEB_XML_SSO blocks

    progressMsg("Writing new web.xml");
    my $WXMLIN;
    my $WXMLOUT;
    my $foundSSOClauseAlready = 0;
    open(WXMLIN, "<", "web.xml") or die "Could not open web.xml for writing.\n";
    open(WXMLOUT, ">", "web.xml.new") or die "Could not open web.xml.new for writing.\n";
    while (<WXMLIN>) {
        #</web-app>
        if (/$WEB_XML_SSO_BEGIN/i) {
            warnMsg("Found SSO section in web.xml already, replacing.\n");
            # already has the SSO stuff in it, replace it
            while (<WXMLIN>) {
                # skip through to the end.
                if (/$WEB_XML_SSO_END/i) {
                    last;
                }
            }
            next;
        }
        if (/\<\/web-app\>/i) {
            progressMsg("Writing web.xml");
            print WXMLOUT $WEB_XML_SSO_BEGIN;
            # if not there just use empty string.
            $g_sso_config->{SSO_CQWEB_OSLC_CONSTRAINT} = getOslcConstraint($g_sso_config->{SSO_CQWEB_OSLC_CONSTRAINT});
            my $webXmlSsoBlock = stringFromTemplateAndHash($WEB_XML_SSO, $g_sso_config);            
            print WXMLOUT $webXmlSsoBlock;
            print WXMLOUT $WEB_XML_SSO_END;
        }
        print WXMLOUT $_;
    }
    close(WXMLIN);
    close(WXMLOUT);
    
    # If already found the clause, just reuse original web.xml
    copy("web.xml.new", "web.xml");
    
    progressMsg("Writing new sso.properties");
    writeSSOProperties();
    
    progressMsg("Writing new web.xml and sso.properties to CQWeb server, this could take a while.");
    runWsAdminFile("putfiles.py");
    
    statusBanner("Done setting up CQ WAS Server for SSO");
}

sub disable_cqweb
{
    statusBanner("Disabling SSO on CQ WAS Server");
    
    #remove these before getting copies from the profile.
    unlink("sso.properties");
    unlink("web.xml");
    
    progressMsg("Getting current web.xml and sso.properties and backing them up");
    getCQWebFilesToEdit();
    
    # make backups
    copyToBackups("sso.properties", "web.xml");
    
    # For web.xml add the $WEB_XML_SSO blocks

    progressMsg("Writing new web.xml");
    my $WXMLIN;
    my $WXMLOUT;
    my $removingSSOClause = 0;
    open(WXMLIN, "<", "web.xml") or die "Could not open web.xml for writing.\n";
    open(WXMLOUT, ">", "web.xml.new") or die "Could not open web.xml.new for writing.\n";
    while (<WXMLIN>) {
        #</web-app>
        if (/$WEB_XML_SSO_BEGIN/i) {
            # already has the SSO stuff in it, don't do anything.
            $removingSSOClause = 1;
            progressMsg("Found SSO section in web.xml, removing.\n");
        }
        if ($removingSSOClause) {
            if (/$WEB_XML_SSO_END/i) {
                $removingSSOClause = 0;        
            }
            next;
        }
        print WXMLOUT $_;
    }
    close(WXMLIN);
    close(WXMLOUT);
    
    # If already found the clause, just reuse original web.xml
    copy("web.xml.new", "web.xml");
    
    progressMsg("Writing new sso.properties");
    writeSSOProperties();
    
    progressMsg("Writing new web.xml and sso.properties to CQWeb server.");
    runWsAdminFile("putfiles.py");
    
    statusBanner("Done disabling up CQ WAS Server for SSO");
}

#TODO - REMOVE - seems like CQWeb already gets installed with ClearQuestUsers mapped to
# "All Authenticated in Trusted Realms"
#-MapRolesToUsers: Mapping users to roles
#Role:  ClearQuestUsers
#Everyone?:  No
#All authenticated?:  No
#Mapped users:
#Mapped groups:
#All authenticated in trusted realms?:  Yes
#Mapped users access ids:
#Mapped groups access ids:
#AdminApp.edit("AppName", '[-MapRolesToUsers [["RegistryClientRole" "no" "no" "uid" "" #"no" "" ""]]]')
#AdminConfig.save()
sub mapAllAuthenticatedUsersToClearQuestUsers
{
    progressMsg("Mapping all authenticated users to ClearQuestUsers");
    #AdminApp.edit('TeamEAR', '[ -MapRolesToUsers [[ ClearQuestUsers AppDeploymentOption.No AppDeploymentOption.No "" "" AppDeploymentOption.Yes "" "" ]]]' ) 
    my @commands = (
        "$AdminApp edit TeamEAR {-MapRolesToUsers {{ClearQuestUsers \"no\" \"no\" \"\" \"\" \"yes\"}}}",
        "$AdminConfig save"
    );
    runWsAdminCommands(\@commands);
    return 
}

sub setApplicationSecurity
{
    my $turnOn = shift;
    progressMsg("Setting application security ". ($turnOn ? "on" : "off"));
    my @commands;
    @commands = (
        "$AdminTask setAdminActiveSecuritySettings { -appSecurityEnabled ".($turnOn ? "true" : "false")."}",
        "$AdminConfig save"
    );
    runWsAdminCommands(\@commands);
    return 
}

sub setup_webserver_oidc
{
    my $enable = shift;
    
    if ($enable) {
        progressMsg("Installing, enabling and configuring the WebSphere OIDC Client EAR.");
        # Install the ear
        my $installScript = findOIDCInstallScript();
        runWsAdminFile(quotePath($installScript), "install ".getWASNode()." ".getWASServerName());
    
        my $setupScript = findOIDCSetupScript();
        my $oidcIdpUrl = $g_sso_config->{SSO_OIDC_IDP_URL};
        my $oidcIdpId = $g_sso_config->{SSO_OIDC_IDP_ID};
        my $oidcIdEndpointPath = $g_sso_config->{SSO_OIDC_IDP_ENDPOINT_PATH} || "endpoint/$oidcIdpId";
        my $oidcIdScope = $g_sso_config->{SSO_OIDC_IDP_SCOPE} || "openid general";
        $oidcIdScope = "\"$oidcIdScope\"";
        my $oidcIdSigAlg = $g_sso_config->{SSO_OIDC_IDP_SIGNATURE_ALG} || "HS256";
        my $oidcIdJwkEndpoint = $g_sso_config->{SSO_OIDC_IDP_JWK_ENDPOINT} || "";
        my $oidcIdIssuerIdentifier = $g_sso_config->{SSO_OIDC_IDP_ISSUER_IDENTIFIER} || "";
        my $oidcIdUserIdentifier = $g_sso_config->{SSO_OIDC_IDP_USER_IDENTIFIER} || "";
        
        my $oidcRpId = $g_sso_config->{SSO_OIDC_RP_ID} || "op";
        my $oidcRpClientId = $g_sso_config->{SSO_OIDC_RP_CLIENTNAME} || "CQWEB";
        my $oidcRpIntercept = $g_sso_config->{SSO_OIDC_RP_INTERCEPT} || "/cqweb.*";
        my $oidcRpSecret = $g_sso_config->{SSO_OIDC_RP_SECRET};
        my $oidcRpProviderNo = $g_sso_config->{SSO_OIDC_RP_PROVIDER} || 1;
        runWsAdminFile(quotePath($setupScript), 
            "config $oidcIdpUrl $oidcIdpId $oidcRpId $oidcRpClientId". 
            " $oidcRpIntercept $oidcRpSecret $oidcRpProviderNo".
            " $oidcIdEndpointPath $oidcIdScope $oidcIdSigAlg $oidcIdJwkEndpoint".
            " $oidcIdIssuerIdentifier $oidcIdUserIdentifier");
    
        my $oidcIdpRealm = $g_sso_config->{SSO_OIDC_IDP_REALM};
        if ($oidcIdpRealm) {
            runWsAdminFile(quotePath($setupScript), "addtrustedrealm $oidcIdpRealm");
        }
    } else {
        progressMsg("SSO is disabled but leaving the WebSphere OIDC Client EAR installed.");
    }
}

# Steps for this are described here:
# http://www.ibm.com/support/knowledgecenter/SSAW57_8.5.5/com.ibm.websphere.nd.doc/ae/twbs_enablesamlsso.html
sub setup_webserver_saml2
{
    my $enable = shift;

    if ($enable) {
        progressMsg("Installing, enabling and configuring the WebSphere SAML2 Client EAR.");
        # Install the SAML Assertion Consumer Service (ACS) ear
        my $installScript = findSAMLACSInstallScript();
        runWsAdminFile(quotePath($installScript), "install ".getWASNode()." ".getWASServerName());
    
        # SAML2 doesn't have 'setup script' like OIDC has, but it does have
        # a wsadmin function, AdminTask.addSAMLTAISSO that we can use.
        #AdminTask.addSAMLTAISSO('-enable true -acsUrl https://<hostname>:<sslport>/samlsps/<any URI pattern string>') 
    
        # A number for identifying which id to use. 
        my $saml2SpId = $g_sso_config->{SSO_SAMP2_SP_ID} || "1";
        my $saml2IdpId = $g_sso_config->{SSO_SAMP2_IDP_ID} || "1";
        # e.g. https://www.myidp.ibm.com/isam/sps/saml20idp/saml20
        my $saml2IdpUrl = $g_sso_config->{SSO_SAML2_IDP_URL};
        # for importing the IDP certificates.
        my $saml2IdpCertAlias = $g_sso_config->{SSO_SAML2_IDP_CERT_ALIAS} || $saml2IdpUrl;
        my $saml2Host = $g_sso_config->{SSO_CQWEB_HOST};
        my $saml2Port = $g_sso_config->{SSO_CQWEB_PORT} || "12443";
        my $acsUrl = "https://$saml2Host:$saml2Port/samlsps/acs";
        # e.g. https://www.myidp.ibm.com/isam/sps/saml20idp/saml20
        # e.g. https://www.myidp.ibm.com/isam/sps/saml20idp/saml20/logininitial?RequestBinding=HTTPPost&PartnerId=https://ADAMSWEB.dev.ratlcore.ibm.com:9450/samlsps/acs&NameIdFormat=Email&Target=https://ADAMSWEB.dev.ratlcore.ibm.com:9450/cqweb 
        my $saml2ErrorPage = $g_sso_config->{SSO_SAML2_ERROR_PAGE_URL} ||
             ("$saml2IdpUrl/logininitial?RequestBinding=HTTPPost&PartnerId=$acsUrl".
              "&NameIdFormat=Email".
              "&Target=https://$saml2Host:$saml2Port/cqweb");
        my $saml2IdpMetaDataFile = $g_sso_config->{SSO_SAML2_IDP_METADATA_FILE};
        if (!$saml2IdpMetaDataFile) {
            errorMsgAndDie("No IDP Metadata file specified. Set this in the configuration file using".
                " the SSO_SAML2_IDP_METADATA_FILE option.");
        }
        # sso_1.idp_1.EntityId
        # needed?
        # my $saml2IdpEntityId = $saml2idpUrl;
    
        # Delete the idp stuff first.
        progressMsg("Deleting SAML2 IdP for ssoId = $saml2SpId and idpId = $saml2IdpId");
        my @commands = (
            "AdminTask.deleteSAMLIdpPartner('-ssoId $saml2SpId -idpId $saml2IdpId -deleteSigningCert true')",
            "AdminConfig.save()"
        );
        runWsAdminCommands(\@commands, $USE_JYTHON );
        
        progressMsg("Deleting existing SAML2 SP for ssoId = $saml2SpId");
        @commands = (
            "AdminTask.deleteSAMLTAISSO('-ssoId $saml2SpId -deleteSigningCert true')",
            "AdminConfig.save()"
        );
        runWsAdminCommands(\@commands, $USE_JYTHON );
        
        progressMsg("Add SAML2 SP for ssoId = $saml2SpId");
        @commands = (
            "AdminTask.addSAMLTAISSO('-ssoId $saml2SpId -enable true ".
                        "-acsUrl https://$saml2Host:$saml2Port/samlsps/acs ".
                        "-idMap idAssertion ".
                        "-errorPage $saml2ErrorPage')",
            "AdminConfig.save()"
        );    
        
        runWsAdminCommands(\@commands, $USE_JYTHON );
        
        progressMsg("Add SAML2 IdP for ssoId = $saml2SpId and idpId = $saml2IdpId");
        @commands =(
            "AdminTask.importSAMLIdpMetadata('-idpMetadataFileName $saml2IdpMetaDataFile".
                                            " -ssoId $saml2SpId -idpId $saml2IdpId ".
                                            " -signingCertAlias $saml2IdpCertAlias')",
            "AdminConfig.save()"
        );
        
        runWsAdminCommands(\@commands, $USE_JYTHON );
    
        # The SP Keys are needed if you want SAML2 communications to use encryption
        my $spKeyStore = $g_sso_config->{SSO_CQWEB_SP_KEY_STORE};
        my $spKeyAlias = $g_sso_config->{SSO_CQWEB_SP_KEY_ALIAS};
        my $spKeyPw = $g_sso_config->{SSO_CQWEB_SP_KEY_PW};
        
        my $customProps = "\"sso_$saml2SpId.sp.trustedAlias=$saml2IdpCertAlias\"";
        if ($spKeyStore) {
            $customProps .= ",";
            $customProps .= "\"sso_$saml2SpId.sp.keyStore=$spKeyStore\"";
        }
        if ($spKeyAlias) {
            $customProps .= ",";
            $customProps .= "\"sso_$saml2SpId.sp.keyAlias=$spKeyAlias\"";
        }
        if ($spKeyPw) {
            $customProps .= ",";
            $customProps .= "\"sso_$saml2SpId.sp.keyPassword=$spKeyPw\"";
        }
                         
    
        # Need sso_1.sp.trustedAlias = the alias we imported the IDP certificate as.
        progressMsg("Add additional custom properties for SAML2 TAI.");
        @commands = (
            "AdminTask.configureInterceptor('[-interceptor com.ibm.ws.security.web.saml.ACSTrustAssociationInterceptor ".
                    "-customProperties [$customProps]]')",
            "AdminConfig.save()"
        );
        runWsAdminCommands(\@commands, $USE_JYTHON );
        
        # Need to tell WebSphere to trust the SAML2 IDP Realm.
        # Use the setting in the configuration file, but if it doesn't exist use
        # the IDP URL, which is the realm for IBM ISAM.
        my $saml2IdpRealm = $g_sso_config->{SSO_SAML2_IDP_REALM} || $saml2IdpUrl;
        progressMsg("Telling WebSphere to trust SAML2 realm: $saml2IdpRealm.");
        @commands = (
            "AdminTask.addTrustedRealms('[-communicationType inbound -realmList $saml2IdpRealm]')",
            "AdminConfig.save()"
        );
        runWsAdminCommands(\@commands, $USE_JYTHON );
    } else {
        progressMsg("SSO is disabled but leaving the WebSphere SAML2 Client EAR installed.");
    }
}

# For WebSphere setup for any SSO types
sub setup_webserver_any
{
    my $enable = shift;
    if (!$enable) {
        # Don't disable SSO on WebSphere. If you do, you can't login to 
        # admin console.  
        # SECJ0154E: SSO Configuration error. FormLogin is configured for 
        # web application isclite but SSO is not enabled in the global 
        # security settings.  SSO must be enabled to use FormLogin.
        return;
    }
    
    # Set SSO Domain and enable SSO
    my $ssoWasRequireSSL = 'true';
    if (defined($g_sso_config->{SSO_WAS_SSO_REQUIRE_SSL}) &&
        ($g_sso_config->{SSO_WAS_SSO_REQUIRE_SSL} eq 'false' ||
         $g_sso_config->{SSO_WAS_SSO_REQUIRE_SSL} eq '0'))
    {
        $ssoWasRequireSSL = 'false';
    }    
    
    my $pMsg = ($enable ? "Enabling" : "Disabling")." SSO";
    if ($enable && $ssoWasRequireSSL) {
        $pMsg .= ", requiring SSL";
    }
    
    my $configureSSOCmd = "-enable ".($enable ? "true" : "false");
    my $ssoWasDomain = $enable ? $g_sso_config->{SSO_WAS_SSO_DOMAIN} : "";
    
    if ($enable) {
        $configureSSOCmd .= " -requiresSSL $ssoWasRequireSSL";
    } 

    if (defined($ssoWasDomain)) {
        $configureSSOCmd .= " -domainName \"$ssoWasDomain\"";
        $pMsg .= ", with SSO domain set to: \"$ssoWasDomain\"";
    }

    progressMsg($pMsg);
    my @commands = (
        "AdminTask.configureSingleSignon('[$configureSSOCmd]')",
        "AdminConfig.save()"
    );
    runWsAdminCommands(\@commands, $USE_JYTHON );
   
    if ($enable) {
        my $saml2Ltpa2Cookie = $g_sso_config->{SSO_WAS_SSO_LTPA2_COOKIE};
        if (defined($saml2Ltpa2Cookie)) {
            progressMsg("Setting LTPA2 cookie name to : \"$saml2Ltpa2Cookie\"");
            @commands = (
                "AdminTask.setAdminActiveSecuritySettings('[-customProperties ".
                    "[\"com.ibm.websphere.security.customSSOCookieName=$saml2Ltpa2Cookie\"]]')",
                "AdminConfig.save()"
            );
            runWsAdminCommands(\@commands, $USE_JYTHON );
        }
    }
}

sub setup_webserver
{
    statusBanner("Doing webserver setup.");
    my $ssoType = $g_sso_config->{SSO_TYPE};
    my $enable = ($g_command eq "enable");
        
    if ($enable) {
        mapAllAuthenticatedUsersToClearQuestUsers();
    }
    # set/unset application security if necessary
    setApplicationSecurity($enable);

    switch ($ssoType) {
        case "OIDC" {
            setup_webserver_oidc($enable);
        }
        case "SAML2" {
            setup_webserver_saml2($enable);
        }
        else {
            errorMsgAndDie("Unknown SSO Type specified: \"$ssoType\"");
        }
    }
    
    # Now do webserver configuration needed for all SSO types.
    setup_webserver_any($enable);
    
    statusBanner("Done doing webserver setup.");
}

sub runWsAdminFile
{
    my $filename = shift;
    my $restOfArgs = join(" ", @_);
    my $wsadmin = getWsAdminCmd();
    runCmd($wsadmin." -profileName $g_CQProfileName -user $g_user -password $g_password -f $filename $restOfArgs");
}

sub runWsAdminCommands
{
    my $commands = shift @_;
    my $useJython = shift @_;
    my $filename = (defined($useJython) && $useJython == $USE_JYTHON) ? "runcommands.py" : "runcommands.jacl";
    open (FH,">",$filename);
    print FH join ("\n", @{$commands});
    logMsg("Writing commands\n". join("\n", @{$commands})."\n");
    close FH;
    my $wsadmincmd = getWsAdminCmd();
    $wsadmincmd .= " -profileName $g_CQProfileName -user $g_user -password $g_password -f $filename";
    if ($useJython) {
        $wsadmincmd .= " -lang jython";
    }
    eval {
        runCmd($wsadmincmd);
    };
    if ($@) {
        errorMsg("Error running commands: $@\n");
    }
    unlink $filename;
}

sub runWsAdminCommandsJython
{
    my ($commands) = @_;
    my $filename = "runcommands.jacl";
    open (FH,">",$filename);
    print FH join ("\n", @{$commands});
    tprint "Writing commands\n". join("\n", @{$commands})."\n";
    close FH;
    my $wsadmin = getWsAdminCmd();
    runCmd($wsadmin." -profileName $g_CQProfileName -user $g_user -password $g_password -f $filename");
    unlink $filename;
}

## cqssosetup.pl unit tests
sub doUnitTests
{
    require Test::Simple;
    import Test::Simple tests=>12;
    
    ok(verifyEnoughForDbSet(("dbset1","user","pw"))) or print "Had enough but it complained I didn't.\n"; 
    ok(!verifyEnoughForDbSet(("dbset1","user"))) or print "Did not have enough but it said I did.\n"; 
    ok(!verifyEnoughForDbSet(("dbset1"))) or print "Did not have enough but it said I did.\n"; 
    ok(scalar(getDbInfo()) == 0) or print "Expected no dbinfo yet.\n";

    my $oslcConstraint = getOslcConstraint();
    ok($oslcConstraint eq "") or print "OSLC Constraint wasn't empty.\n";

    my $configConstraint = "<url-pattern>test</url-pattern>";
    $oslcConstraint = getOslcConstraint($configConstraint);
    ok($oslcConstraint eq $configConstraint) or print "OSLC Constraint should have been:\n$configConstraint\nactually: \n$oslcConstraint\n";

    $g_sso_config->{SSO_CQ_DATABASE_INFO} = "dbset1,user,pw";
    ok(scalar(getDbInfo()) == 3) or print "Expected 3 dbinfo elements but had : ".Dumper(getDbInfo())."\n";
    
    my @dbsets = getDbSets();
    ok(scalar(@dbsets) == 1) or print "Wrong number of elements in dbsets. Expected 1, got ".scalar(@dbsets)."\n";

    $g_sso_config->{SSO_CQ_DATABASE_INFO} = "dbset1,user,pw,dbset2,user2,pw2,dbset3,user3,pw3";
    @dbsets = getDbSets();
    ok(scalar(@dbsets) == 3) or print "Wrong number of elements in dbsets. Expected 3, got ".scalar(@dbsets)."\n";
        
    # should still be empty since JAS/OIDC not enabled
    $oslcConstraint = getOslcConstraint();
    ok($oslcConstraint eq "") or print "OSLC Constraint wasn't empty.\n";
    
    #now its enabled, should have the discovery elements for each dbset
    $g_sso_config->{SSO_USES_JAS_OIDC} = "true";
    $oslcConstraint = getOslcConstraint();
    my $expectedConstraint = "<url-pattern>/oslc/repo/dbset1/discovery</url-pattern>".
                            "<url-pattern>/oslc/repo/dbset2/discovery</url-pattern>".
                            "<url-pattern>/oslc/repo/dbset3/discovery</url-pattern>";
    ok($oslcConstraint eq $expectedConstraint) or print "OSLC Constraint did not match expected.\n".
                "Expected: $expectedConstraint\nActual  : $oslcConstraint\n";    

    $expectedConstraint = $configConstraint.$expectedConstraint;
    $oslcConstraint = getOslcConstraint($configConstraint);
    ok($oslcConstraint eq $expectedConstraint) or print "OSLC Constraint did not match expected.\n".
                "Expected: $expectedConstraint\nActual  : $oslcConstraint\n";    
    
}

sub doUsage
{
	print "Usage: cqperl cqssosetup.pl help\n";
	print "  Version: $cqssosetup_version.\n";
	print "     Print this help message.\n";
	print "  or:  cqperl cqssosetup.pl enable [-config <configfile>] [-nocqrpc] [-nomasterdbs]\n";
	print "                                   [-nocqweb] [-nowebserver] [-norestart] [-nopost]\n";
	print "     Enable CQ Web for SSO. If a '-noXXXX' option is present then that step is\n";
	print "     skipped (see below for details). This script can be called multiple times.\n";
	print "     and the full steps can be re-run multiple times safely. The only reason to\n";
	print "     to skip any steps is if you know the step can be skipped and you want to save\n";
	print "     time. It is recommended you never skip any steps. For example if you want to\n";
	print "     SSO password, you can change it in the configuration file and re-run this command.\n";
	print "     Description of arguments:\n";	
	print "         -config <configfile>\n";	
	print "     Location of the configuration file used for the command. Default: ssoconfig.txt\n";	
	print "         -nocqrpc\n";
    print "     Do not set the SSO password used by cqrpc. Once set you do not need to set it\n";
    print "     again unless the SSO password changes.\n";	
	print "         -nomasterdbs\n";
    print "     Do not set up the ClearQuest databases for SSO.\n";	
	print "         -nocqweb\n";
    print "     Do not add the <security-constraint> to the ClearQuest Web application (web.xml)\n";
    print "     to add security.\n";	
	print "         -nowebserver\n";
    print "     Do not add and configure the EAR necessary to suppor the desired SSO technology.\n";
    print "     For OpenID Connect, the WebSphereOIDCRP.ear is added to the profile.\n";	
	print "     For SAML2, the WebSphereSamlSP.ear is added to the profile.\n";	
	print "     These EARs are provided by WebSphere Application Server.";
	print "         -norestart\n";
    print "     Do not restart the WebSphere profile.\n";	
	print "         -nopost\n";
    print "     After setup is complete, the script prints out directions and support files to\n";	
    print "     be used to complete the SSO setup. This argument skips that step.\n";	
	print "  or:  cqperl cqssosetup.pl disable \n";
	print "     Disable CQ Web for SSO. Application security is disabled, sso properties is changed to turn off SSO.\n";
	print "  or:  cqperl cqssosetup.pl restart \n";
	print "     Utility for stopping/restarting CQ Web server.\n";
	print ""
}

$g_command = shift @ARGV;

if ($g_command eq "enable") {
    sso_enable();
}
elsif ($g_command eq "disable") {
    sso_disable();
}
elsif ($g_command eq "restart") {
    sso_restart();
} 
elsif (!$g_command || $g_command eq "help") {
    doUsage();
} 
elsif ($g_command eq "unittest") {
    doUnitTests();
}
else {
    errorMsg("Unrecognized command.");
    doUsage();
} 

close($LOGFILE);

1;
