#------------------------------------------------------------------------------
# OIDC Client application deployment/undeployment script
# wsadmin.sh -f setupoidc.py (oidIdPUrl) (oidIdIdentifer) (oidcRpIdentifier) (oidcRpClientName) (oidcRpIntercept) (oidcRpSecret) (oidcRpProviderNo) 
#                    [oidcEndpointPath [oidcScope [oidcSignature [oidcSignatureAlg [oidcJwkEndpoint [oidcIssuerIdentifer [oidcUserIdentifier]]]]]]
#                    [name=value ...]
#               Setup and configure this WebSphere server as an OIDC Relying Party.
#               Arguments:
#               oidcIdPUrl - OIDC ID providers URL including the ID, and trailing slash, e.g.,
#                           "https://www.myidp.com:9443/oidc/
#               oidIdIdentifer - The OIDC ID providers identifier, e.g. "MY_OP"
#               oidcRpIdentifier - The identifier for the relying party, e.g. "libop"
#                                   Used for registering an RP client on an IDP.
#               oidcRpClientName - The client name for the relying party, e.g. "MY_RP"
#               oidcRpIntercept - The intercept filter for the RP, e.g. "/.*"
#               oidcRpSecret - A secret word used for establishing trust between OIDC provider and client.
#               oidcRpProviderNo - a number that allows you to register multiple 
#                            OIDC ID providers on the webserver, e.g. "1"
#               oidcEndpointPath - Path to add to oidcIdPUrl to get to the OIDC endpoints
#               oidcScope - The scope to request from the OIDC IDP
#               oidcSignatureAlg - The signature algorithm to use, HS256 (default) or RS256
#               oidcJwkEndpoint - The path to add after oidcEndpointPath to reach the JWK endpoint URL
#               oidcIssuerIdentifer - The URL to the issuer identifier (sets the issuerIdentifier in TAI)
#               oidcUserIdentifier - The attribute from the users profile to use for the principal name
#               [optional additional name=value configurations for the TAI custom properties.]
#
# wsadmin.sh -f setupoidc.py setproperty name=value ...
#               Set or add custom properties to the OIDC Relying Party TAI.
#
# wsadmin.sh -f setupoidc.py.py unconfig
#
# wsadmin.sh -f setupoidc.py.py addtrustedrealm (trustedrealm)
#               Add a trusted realm to WebSphere. For example if the IDP realm is "MyIDPRealm"
#               you need to add it as a trusted realm for WebSphere to trust it.
#               
#------------------------------------------------------------------------------
import sys

#------------------------------------------------------------------------------
# Print script usage
#------------------------------------------------------------------------------
def printUsage():
	print "Usage: wsadmin -f <WAS_HOME>/bin/setupoidc.py.py help"
	print "  or:  wsadmin -f <WAS_HOME>/bin/setupoidc.py.py config "
	print "             (oidIdPUrl) (oidIdIdentifer) (oidcRpIdentifier) (oidcRpClientName) (oidcRpIntercept) (oidcRpSecret) (oidcRpProviderNo) "
	print "             [oidcEndpointPath [oidcScope [oidcSignatureAlg [oidcJwkEndpoint [oidcIssuerIdentifer "
	print "             [oidcUserIdentifier [(name)=(value) ... ]]]]]]]"
	print "  or:  wsadmin -f <WAS_HOME>/bin/setupoidc.py.py setproperty (name)=(value) ... "
	print "  or:  wsadmin -f <WAS_HOME>/bin/setupoidc.py.py unconfig"
	print ""

#------------------------------------------------------------------------------
# Unconfigure the application
#------------------------------------------------------------------------------
def doUnconfig():
    failOnError = "false"
    print "*** Unconfiguring the OIDC RP interceptor..."
    try:
        AdminTask.unconfigureInterceptor("[-interceptor com.ibm.ws.security.oidc.client.RelyingParty ]")
        # AdminTask.configureTrustAssociation("-enable false")
        AdminConfig.save()
        print "*** Done unconfiguring the OIDC RP interceptor..."
    except:
    	print '!!! Failed to unconfigure the OIDC RP interceptor ', sys.exc_info()
    failOnError = "true"

#------------------------------------------------------------------------------
# Configure a custom property for the OpenIDC RP Trust Association Intercaptor (TAI)
#------------------------------------------------------------------------------
def setCustomPropertyForProvider(providerno, name, value):
    newName = "provider_"+providerno+"." + name;
    nameValue = newName + "=" + value
    setCustomProperty(nameValue)
     
#------------------------------------------------------------------------------
# Configure a custom property for the OpenIDC RP Trust Association Intercaptor (TAI)
# Accepts the namevalue as an "name=value" string
#------------------------------------------------------------------------------
def setCustomProperty(pNameValue): 
    failOnError = "false"
    #"AdminTask.configureInterceptor('[-interceptor com.ibm.ws.security.oidc.client.RelyingParty -customProperties ]') 
    #configArg = '[-interceptor com.ibm.ws.security.oidc.client.RelyingParty -customProperties ["test=blah"]]'
    configArg = '[-interceptor com.ibm.ws.security.oidc.client.RelyingParty -customProperties [\"'+pNameValue+'\"]]'
    print "*** Setting custom propert: " + pNameValue
    try:
    	AdminTask.configureInterceptor(configArg)
    except:
    	print 'Failed to configure the OIDC RP TAI interceptor with custom property ', sys.exc_info()	    
    #AdminTask.configureInterceptor('[-interceptor com.ibm.ws.security.oidc.client.RelyingParty -customProperties [\"'+pNameValue+'\"]]')
    failOnError = "true"
#------------------------------------------------------------------------------
# Configure custom properties from argv for the OpenID RP Trust Association Intercaptor (TAI)
#------------------------------------------------------------------------------
def setCustomPropertiesFromArgv(argv, startAt):
	nowAt=startAt
        stopAt=len(argv)
	while nowAt < stopAt:
		setCustomProperty(argv[nowAt])
		nowAt += 1
	# Save the config after setting all custom properties requested
	AdminConfig.save()
	

#------------------------------------------------------------------------------
# given a URL, get the port number (assumes port 443 by default for https URLS)
#------------------------------------------------------------------------------
def getPortFromUrl(oidcIdPUrl):
	from java.net import URL
	url = URL(oidcIdPUrl)
	port = url.getPort()
	if port==-1 and	url.getProtocol()=="https":
		port=443
	return str(port)

#------------------------------------------------------------------------------
# given a URL, get the host string
#------------------------------------------------------------------------------
def getHostFromUrl(oidcIdPUrl):
	from java.net import URL
	url = URL(oidcIdPUrl)
	host = url.getHost()
	return host

#------------------------------------------------------------------------------
# Get the public signer certificate for a given HTTPS URL, and configure it into WebSphere's trust store
#------------------------------------------------------------------------------
def getOidOpSignerCert(oidcIdPUrl):
    # 
    failOnError = "false"
    cell=AdminControl.getCell()
    node=AdminControl.getNode()
    scope='(cell):%s:(node):%s' % (cell, node)
    oidcidphost=getHostFromUrl(oidcIdPUrl)
    oidcidpport=getPortFromUrl(oidcIdPUrl)
    #
    alias=oidcIdPUrl.lower()
    #
    # delete existing certificate
    try:
        print "*** Trying to delete existing certificate from scope:%s alias:%s." % (scope, alias)
        myArgList = '[-keyStoreName NodeDefaultTrustStore -keyStoreScope %s -certificateAlias %s ]' % (scope, alias)
        print "Trying AdminTask.deleteSignerCertificate('%s')" % myArgList
        AdminTask.deleteSignerCertificate(myArgList)
        print "*** Done deleting existing certificate."
    except:
    	print 'Failed to delete OIDC IDP certificate ', sys.exc_info()[1]
    # import OID OP certificate
    try:
    	print '*** Retrieving signer from OID IDP URL: '+ oidcIdPUrl + '.'
    	myArgList = '[-keyStoreName NodeDefaultTrustStore -keyStoreScope %s -host %s -port %s -certificateAlias %s -sslConfigName NodeDefaultSSLSettings -sslConfigScopeName %s ]' % (scope, oidcidphost, oidcidpport, alias, scope);
    	print "Trying AdminTask.retrieveSignerFromPort('%s')" % (myArgList)
    	AdminTask.retrieveSignerFromPort(myArgList)
        print '*** Stored signer in keyStoreScope:' + scope + ', alias:' + alias + '.'
        
    except:
    	print '!!! Failed to add the OIDC IDP certificate ', sys.exc_info()
    AdminConfig.save()
    #
    certlist=AdminTask.listSignerCertificates('[-keyStoreName NodeDefaultTrustStore -keyStoreScope %s ]' % scope) 
    if certlist.find(alias) == -1:
    	print '!!! Failed to import the certficate from OIDC IP server.. may need to correct it manually'
    else:
    	print '*** Successfully imported OIDC IP server certificate'
    failOnError = "true"

#------------------------------------------------------------------------------
# Initial Configuration for the OpenID RP Trust Association Interceptor (TAI)
#------------------------------------------------------------------------------
def doConfigure(oidcIdPUrl, idpidentifier, rpidentifier, clientid, intercept, secret, providerNo, endpointPath, scope, sigAlg, jwkEndpoint, issuerIdentifier, userIdentifier):
	getOidOpSignerCert(oidcIdPUrl)
	doUnconfig()
	print '*** Enabling trust association'
	AdminTask.configureTrustAssociation("-enable true")
	print '*** Enabled trust association'
	# TODO - don't overwrite, add to the end?
	print '*** Setting InvokeTAIbeforeSSO'
	AdminTask.setAdminActiveSecuritySettings('-customProperties  \'"com.ibm.websphere.security.InvokeTAIbeforeSSO=com.ibm.ws.security.oidc.client.RelyingParty"\'')
	print '*** Done setting InvokeTAIbeforeSSO'
	
	oidcIdPUrlBase = oidcIdPUrl + endpointPath + "/"
	
	print '*** Setting Custom properies for OIDC RP TAI'
	setCustomPropertyForProvider(providerNo, "interceptedPathFilter", intercept)
	setCustomPropertyForProvider(providerNo, "identifier", rpidentifier)
	setCustomPropertyForProvider(providerNo, "clientId", clientid)
	setCustomPropertyForProvider(providerNo, "clientSecret", secret)
	setCustomPropertyForProvider(providerNo, "authorizeEndpointUrl", oidcIdPUrlBase + "authorize")
	setCustomPropertyForProvider(providerNo, "tokenEndpointUrl", oidcIdPUrlBase + "token")
	setCustomPropertyForProvider(providerNo, "introspectEndpointUrl", oidcIdPUrlBase + "introspect")
	setCustomPropertyForProvider(providerNo, "scope", scope)
	setCustomPropertyForProvider(providerNo, "signatureAlgorithm", sigAlg)
	
	if jwkEndpoint != "":
	    setCustomPropertyForProvider(providerNo, "jwkEndpointUrl", oidcIdPUrlBase + jwkEndpoint)
		
	if issuerIdentifier != "":
	    setCustomPropertyForProvider(providerNo, "issuerIdentifier", issuerIdentifier)
		
	if userIdentifier != "":
	    setCustomPropertyForProvider(providerNo, "userIdentifier", userIdentifier)
		
	AdminConfig.save()

def addTrustedRealm(trustedRealm):
    AdminTask.addTrustedRealms('[-communicationType inbound -realmList %s]' % trustedRealm)
    AdminConfig.save();

#------------------------------------------------------------------------------
# runscript - run the script as a subroutine
#------------------------------------------------------------------------------
def runscript(args):
  failOnError = "true"

  mode = args[0]
  if mode == "help":
  	printUsage()
  else:
    if mode == "config":
        if len(args) > 7:
            # Optional params
            endpointPath = "endpoint"
            sigAlg = "HS256"
            jwkEndpoint = ""
            issuerIdentifier = ""
            userIdentifier = ""
            scope = "openid general"

            argNo = 1
            oidcIdPUrl = args[argNo]
            argNo += 1
            idpidentifer = args[argNo]
            argNo += 1
            rpidentifer = args[argNo]
            argNo += 1
            clientid = args[argNo]
            argNo += 1
            intercept = args[argNo]
            argNo += 1
            secret = args[argNo]
            argNo += 1
            providerNo = args[argNo]
            argNo += 1
            if len(args) > 8:
                endpointPath = args[argNo]
                argNo += 1 
            if len(args) > 9:
                scope = args[argNo]
                argNo += 1 
            if len(args) > 10:
                sigAlg = args[argNo]
                argNo += 1 
            if len(args) > 11:
                jwkEndpoint = args[argNo]
                argNo += 1 
            if len(args) > 12:
                issuerIdentifier = args[argNo]
                argNo += 1 
            if len(args) > 13:
                userIdentifier = args[argNo]
                argNo += 1 
 
            doConfigure(oidcIdPUrl, idpidentifer, rpidentifer, clientid, intercept, secret, providerNo, endpointPath, scope, sigAlg, jwkEndpoint, issuerIdentifier, userIdentifier)
            setCustomPropertiesFromArgv(args,argNo)
        else:
            sys.stderr.write("!!! Invalid number of arguments for config operation\n")
            printUsage()
            sys.exit(101)
    elif mode == "setproperty":
        if len(args) > 1 :
            setCustomPropertiesFromArgv(args,1)
        else:
            sys.stderr.write("!!! Invalid number of arguments for setproperty operation\n")
            printUsage()
            sys.exit(101)
    elif mode == "unconfig":
        if len(args) == 1:
            doUnconfig()
        else:
            sys.stderr.write("!!! Invalid number of arguments for unconfig operation\n")
            printUsage()
            sys.exit(101)
    elif mode == "addtrustedrealm":
        if len(args) == 2:
            addTrustedRealm(args[1]);
        else:
            sys.stderr.write("!!! Invalid number of arguments for addtrustedrealm operation\n")
            printUsage()
            sys.exit(101)        
    else:
        sys.stderr.write("Invalid operation\n")
        printUsage()
        sys.exit(101)
  
#------------------------------------------------------------------------------
# Main entry point
#------------------------------------------------------------------------------
if len(sys.argv) > 0:
	runscript(sys.argv)
else:
   	printUsage()


